/* ############################# MPC License ############################## */
/* # Wed Nov 19 15:19:19 CET 2008                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # IDDN.FR.001.230040.000.S.P.2007.000.10000                            # */
/* # This file is part of the MPC Runtime.                                # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - PERACHE Marc marc.perache@cea.fr                                 # */
/* #                                                                      # */
/* ######################################################################## */


#define _GNU_SOURCE
#include "sctk_config.h"
#include "sctk_alloc.h"
#include <sctk_debug.h>
#include <sctk_tls.h>

#if defined(SCTK_USE_TLS)
__thread void *sctk_hierarchical_tls = NULL;
#endif

#if defined(SCTK_USE_TLS) && defined(Linux_SYS)
#include <link.h>
#include <stdlib.h>
#include <stdio.h>

__thread unsigned long p_memsz;
__thread void *p_vaddr;
__thread size_t current_module;
__thread size_t total_module;

static int
callback (struct dl_phdr_info *info, size_t size, void *data)
{
  int j;


  for (j = 0; j < info->dlpi_phnum; j++)
    {
      if (info->dlpi_phdr[j].p_type == PT_TLS)
	{
	  total_module++;
	  if (total_module == current_module)
	    {
	      p_memsz = info->dlpi_phdr[j].p_memsz;
	      p_vaddr =
		(void *) ((char *) info->dlpi_addr +
			  info->dlpi_phdr[j].p_vaddr);
	      sctk_nodebug ("TLS in %d %p size %lu start %p", j,
			    info->dlpi_addr, p_memsz, p_vaddr);
	      return 1;
	    }
	  else
	    {
	      sctk_nodebug ("TLS SKIP in %d %p %d != %d", j, info->dlpi_addr,
			    total_module, current_module);
	    }
	}
    }
  return 0;
}


typedef struct
{
  unsigned long int ti_module;
  unsigned long int ti_offset;
} tls_index;

typedef char *sctk_tls_module_t;

typedef struct
{
  volatile int size;
  volatile sctk_tls_module_t *volatile modules;
  sctk_spinlock_t lock;
} tls_level;

static inline void
sctk_tls_init_level (tls_level * level)
{
  level->size = 0;
  level->modules = NULL;
  level->lock = 0;
}
static inline void
sctk_tls_write_level (tls_level * level)
{
  sctk_spinlock_lock (&(level->lock));
}
static inline void
sctk_tls_unlock_level (tls_level * level)
{
  sctk_spinlock_unlock (&(level->lock));
}

static inline size_t
sctk_determine_module_size (size_t m)
{
  return p_memsz;
}

static inline void
sctk_init_module (size_t m, char *module, size_t size)
{
  memcpy (module, p_vaddr, size);
}

char *
sctk_alloc_module (size_t m, tls_level * tls_level)
{
  char *tls_module;
  tls_module = (char *) tls_level->modules[m - 1];
  if (tls_module == NULL)
    {
      size_t module_size;

      current_module = m;
      total_module = 0;
      dl_iterate_phdr (callback, NULL);

      module_size = sctk_determine_module_size (m);
      tls_module = sctk_malloc (module_size);
      sctk_init_module (m, tls_module, module_size);
      tls_level->modules[m - 1] = tls_module;
      sctk_nodebug ("Init module %d", m);
    }
  return tls_module;
}

void *
__sctk__tls_get_addr__generic_scope (tls_index * tmp, size_t m, size_t offset,
				     sctk_tls_scope_t scope)
{
  void *res;
  tls_level **tls;
  tls_level *tls_level;
  char *tls_module;


  if (tmp != NULL)
    {
      m = tmp->ti_module;
      offset = tmp->ti_offset;
    }

  sctk_nodebug ("Module %lu Offset %lu", m, offset);

  tls = sctk_hierarchical_tls;
  assume (tls != NULL);

  tls_level = tls[scope];

  if (tls_level->size < m)
    {
      sctk_tls_write_level (tls_level);

      if (tls_level->size < m)
	{
	  size_t i;
	  tls_level->modules =
	    sctk_realloc ((void *) (tls_level->modules), m * sizeof (char *));
	  for (i = tls_level->size; i < m; i++)
	    {
	      tls_level->modules[i] = NULL;
	      sctk_alloc_module (i, tls_level);
	    }
	  sctk_nodebug ("Init modules to size %ld->%ld", tls_level->size, m);
	  tls_level->size = m;
	}

      tls_module = sctk_alloc_module (m, tls_level);
      sctk_tls_unlock_level (tls_level);

      res = tls_module + offset;
    }
  else
    {
      tls_module = (char *) tls_level->modules[m - 1];
      res = tls_module + offset;
    }

  return res;
}


static void
sctk_tls_create ()
{
  tls_level **tls;
  int i;
  tls = sctk_malloc (sctk_tls_max_scope * sizeof (tls_level *));

  for (i = 0; i < sctk_tls_max_scope; i++)
    {
      tls[i] = sctk_malloc (sizeof (tls_level));
      sctk_tls_init_level (tls[i]);
    }

  sctk_nodebug ("Init tls_hierarchical root %p", tls);
  sctk_hierarchical_tls = tls;
}

void
sctk_tls_duplicate (void **new)
{
  tls_level *tls;
  tls_level *new_tls;
  int i;

  tls = sctk_hierarchical_tls;
  if (tls == NULL)
    {
      sctk_tls_create ();
    }

  new_tls = sctk_malloc (sctk_tls_max_scope * sizeof (tls_level));
  *new = new_tls;

  tls = sctk_hierarchical_tls;
  sctk_nodebug ("Duplicate %p->%p", tls, new_tls);

  for (i = 0; i < sctk_tls_max_scope; i++)
    {
      new_tls[i] = tls[i];
    }

}

void
sctk_tls_keep (int *scopes)
{
  tls_level **tls;
  int i;
  tls = sctk_hierarchical_tls;

  assume (tls != NULL);

  for (i = 0; i < sctk_tls_max_scope; i++)
    {
      if (scopes[i] == 0)
	{
	  sctk_nodebug ("Remove %d in %p", i, tls);
	  tls[i] = sctk_malloc (sizeof (tls_level));
	  sctk_tls_init_level (tls[i]);
	}
    }
}

void
sctk_tls_delete ()
{
#warning "Pas de liberation des tls"
}


#if defined(SCTK_i686_ARCH_SCTK) || defined (SCTK_x86_64_ARCH_SCTK)


void *
__sctk__tls_get_addr__task_scope (tls_index * tmp)
{
  void *res;
  res = __sctk__tls_get_addr__generic_scope (tmp, 0, 0, sctk_tls_task_scope);
  return res;
}

void *
__sctk__tls_get_addr__thread_scope (tls_index * tmp)
{
  void *res;
  res =
    __sctk__tls_get_addr__generic_scope (tmp, 0, 0, sctk_tls_thread_scope);
  return res;
}

void *
__sctk__tls_get_addr__process_scope (tls_index * tmp)
{
  void *res;
  res =
    __sctk__tls_get_addr__generic_scope (tmp, 0, 0, sctk_tls_process_scope);
  return res;
}

#elif defined (SCTK_ia64_ARCH_SCTK)

void *
__sctk__tls_get_addr__task_scope (size_t m, size_t offset)
{
  void *res;
  res =
    __sctk__tls_get_addr__generic_scope (NULL, m, offset,
					 sctk_tls_task_scope);
  return res;
}

void *
__sctk__tls_get_addr__thread_scope (size_t m, size_t offset)
{
  void *res;
  res =
    __sctk__tls_get_addr__generic_scope (NULL, m, offset,
					 sctk_tls_thread_scope);
  return res;
}

void *
__sctk__tls_get_addr__process_scope (size_t m, size_t offset)
{
  void *res;
  res =
    __sctk__tls_get_addr__generic_scope (NULL, m, offset,
					 sctk_tls_process_scope);
  return res;
}
#else
#error "Architecture not available"
#endif
#else
#warning "Experimental TLS support desactivated"

void
sctk_tls_create ()
{
}

void
sctk_tls_duplicate (void **new)
{
  *new = NULL;
}

void
sctk_tls_keep (int *scopes)
{
}

void
sctk_tls_delete ()
{
}

#endif
